@@ -14,14 +14,7 @@ import requests
 from urllib3.exceptions import HTTPError, PoolError
 from requests.exceptions import RequestException
 
-if sys.version_info[0] >= 3:
-    string_types = str
-else:
-    string_types = basestring
-    warnings.warn('Module pdpyras will no longer support Python 2.7 as of '
-        'June 21, 2021.')
-
-__version__ = '4.2.1'
+__version__ = '4.5.1'
 
 # These are API resource endpoints/methods for which multi-update is supported
 VALID_MULTI_UPDATE_PATHS = [
@@ -32,7 +25,7 @@ VALID_MULTI_UPDATE_PATHS = [
 
 ITERATION_LIMIT = 1e4
 
-TIMEOUT = 5
+TIMEOUT = 60
 
 #########################
 ### UTILITY FUNCTIONS ###
@@ -43,7 +36,7 @@ def auto_json(method):
     Function decorator that makes methods return the full response JSON
     """
     def call(self, path, **kw):
-        response = raise_on_error(method(self, path, **pass_kw))
+        response = raise_on_error(method(self, path, **kw))
         return try_decoding(response)
     return call
 
@@ -77,17 +70,20 @@ def raise_on_error(r):
     :returns: The response object, if its status was success
     :rtype: `requests.Response`_
     """
-    if r.ok:
-        return r
+    received_http_response = bool(r.status_code)
+    if received_http_response:
+        if r.ok:
+            return r
+        else:
+            raise PDHTTPError("%s %s: API responded with non-success status "
+                "(%d): %s" % (
+                    r.request.method.upper(),
+                    r.request.url.replace('https://api.pagerduty.com', ''),
+                    r.status_code,
+                    r.text[:99]
+                ), r)
     else:
-        raise PDClientError("%s %s: API responded with non-success status "
-            "(%d): %s" % (
-                r.request.method.upper(),
-                r.request.url.replace('https://api.pagerduty.com', ''),
-                r.status_code,
-                r.text[:99]
-            ), response=r
-        )
+        raise PDClientError("Network or unknown error: "+str(r))
 
 def resource_envelope(method):
     """
@@ -100,7 +96,8 @@ def resource_envelope(method):
 
     Methods using this decorator will raise a :class:`PDClientError` with its
     ``response`` property being being the `requests.Response`_ object in the
-    case of any error, so that the implementer can access it by catching the
+    case of any error (as of version 4.2 this is subclassed as
+    :class:`PDHTTPError`), so that the implementer can access it by catching the
     exception, and thus design their own custom logic around different types of
     error responses.
 
@@ -148,7 +145,7 @@ def resource_envelope(method):
         r = raise_on_error(method(self, path, **pass_kw))
         # Now let's try to unpack...
         response_obj = try_decoding(r)
-        # Get the encapsulated object
+        # Get the wrapped entity
         if envelope_name not in response_obj:
             raise PDClientError("Cannot extract object; expected top-level "
                 "property \"%s\", but could not find it in the response "
@@ -268,8 +265,7 @@ def try_decoding(r):
     try:
         return r.json()
     except ValueError as e:
-        raise PDClientError("API responded with invalid JSON: "+r.text[:99],
-            response=r)
+        raise PDHTTPError("API responded with invalid JSON: "+r.text[:99], r)
 
 ###############
 ### CLASSES ###
@@ -370,6 +366,12 @@ class PDSession(requests.Session):
     and request will increase by a factor of this amount.
     """
 
+    timeout = TIMEOUT
+    """
+    This is the value sent to `requests.Session.request`_ as the ``timeout``
+    parameter that determines the TCP read timeout.
+    """
+
     url = ""
 
     def __init__(self, api_key, name=None):
@@ -384,7 +386,7 @@ class PDSession(requests.Session):
         self.parent = super(PDSession, self)
         self.parent.__init__()
         self.api_key = api_key
-        if isinstance(name, string_types) and name:
+        if isinstance(name, str) and name:
             my_name = name
         else:
             my_name = self.trunc_key
@@ -411,39 +413,12 @@ class PDSession(requests.Session):
 
     @api_key.setter
     def api_key(self, api_key):
-        if not (isinstance(api_key, string_types) and api_key):
+        if not (isinstance(api_key, str) and api_key):
             raise ValueError("API credential must be a non-empty string.")
         self._api_key = api_key
         self.headers.update(self.auth_header)
         self.after_set_api_key()
 
-    @property
-    def api_key_access(self):
-        """
-        Memoized API key access type getter.
-
-        Will be "user" if the API key is a user-level token (all users should
-        have permission to create an API key with the same permissions as they
-        have in the PagerDuty web UI).
-
-        If the API key in use is an account-level API token (as only a global
-        administrator user can create), this property will be "account".
-        """
-        if not hasattr(self, '_api_key_access') or self._api_key_access is None:
-            response = self.get('/users/me')
-            if response.status_code == 400:
-                message = try_decoding(response).get('error', '')
-                if 'account-level access token' in message:
-                    self._api_key_access = 'account'
-                else:
-                    self._api_key_access = None
-                    self.log.error("Failed to obtain API key access level; "
-                        "the API did not respond as expected.")
-                    self.log.debug("Body = %s", response.text[:99])
-            else:
-                self._api_key_access = 'user'
-        return self._api_key_access
-
     @property
     def auth_header(self):
         """
@@ -454,20 +429,51 @@ class PDSession(requests.Session):
     def cooldown_factor(self):
         return self.sleep_timer_base*(1+self.stagger_cooldown*random())
 
+    def normalize_params(self, params):
+        """
+        Modify the user-supplied parameters.
+
+        Current behavior:
+
+        * If a parameter's value is of type list, and the parameter name does
+          not already end in "[]", then the square brackets are appended to keep
+          in line with the requirement that all set filters' parameter names end
+          in "[]".
+        """
+        updated_params = {}
+        for param, value in params.items():
+            if type(value) is list and not param.endswith('[]'):
+                updated_params[param+'[]'] = value
+            else:
+                updated_params[param] = value
+        return updated_params
+
+    def normalize_url(self, url):
+        """Compose the URL whether it is a path or an already-complete URL"""
+        if url.startswith(self.url) or not self.url:
+            return url
+        else:
+            return self.url + "/" + url.lstrip('/')
+
     def postprocess(self, response):
         """
         Perform supplemental actions immediately after receiving a response.
         """
         pass
 
-    def prepare_headers(self, method):
+    def prepare_headers(self, method, user_headers={}):
         """
         Append special additional per-request headers.
 
         :param method:
             The HTTP method, in upper case.
+        :param user_headers:
+            Headers that can be specified to override default values.
         """
-        return self.headers
+        headers = deepcopy(self.headers)
+        if user_headers:
+            headers.update(user_headers)
+        return headers
 
     def request(self, method, url, **kwargs):
         """
@@ -479,9 +485,8 @@ class PDSession(requests.Session):
         :param url:
             The path/URL to request. If it does not start with the base URL, the
             base URL will be prepended.
-        :param \*\*kwargs:
-            Additional keyword arguments to pass to `requests.Session.request
-            <https://2.python-requests.org/en/master/api/#requests.Session.request>`_
+        :param \\*\\*kwargs:
+            Additional keyword arguments to pass to `requests.Session.request`_
         :type method: str
         :type url: str
         :returns: the HTTP response object
@@ -496,16 +501,22 @@ class PDSession(requests.Session):
                 "Method %s not supported by this API. Permitted methods: %s"%(
                     method, ', '.join(self.permitted_methods)))
         req_kw = deepcopy(kwargs)
-        my_headers = self.prepare_headers(method)
-        # Merge, but do not replace, any headers specified in keyword arguments:
-        if 'headers' in kwargs:
-            my_headers.update(kwargs['headers'])
-        req_kw.update({'headers': my_headers, 'stream': False, 'timeout': TIMEOUT})
-        # Compose/normalize URL whether or not path is already a complete URL
-        if url.startswith(self.url) or not self.url:
-            my_url = url
-        else:
-            my_url = self.url + "/" + url.lstrip('/')
+
+        # Add in any headers specified in keyword arguments:
+        headers = kwargs.get('headers', {})
+        req_kw.update({
+            'headers': self.prepare_headers(method, user_headers=headers),
+            'stream': False,
+            'timeout': self.timeout
+        })
+
+        # Special changes to user-supplied parameters, for convenience
+        if 'params' in kwargs and kwargs['params']:
+            req_kw['params'] = self.normalize_params(kwargs['params'])
+
+        # Compose the full URL:
+        my_url = self.normalize_url(url)
+
         # Make the request (and repeat w/cooldown if the rate limit is reached):
         while True:
             try:
@@ -516,7 +527,7 @@ class PDSession(requests.Session):
                 if network_attempts > self.max_network_attempts:
                     raise PDClientError("Non-transient network error; exceeded "
                         "maximum number of attempts (%d) to connect to the "
-                        "API"%self.max_network_attempts)
+                        "API."%self.max_network_attempts)
                 sleep_timer *= self.cooldown_factor()
                 self.log.debug("HTTP or network error: %s: %s; retrying in %g "
                     "seconds.", e.__class__.__name__, e, sleep_timer)
@@ -537,24 +548,24 @@ class PDSession(requests.Session):
                             "status %d.", self.retry[status], status)
                         return response
                     http_attempts[status] = 1 + http_attempts.get(status, 0)
-                sleep_timer *= self.sleep_timer_base
+                sleep_timer *= self.cooldown_factor()
                 self.log.debug("HTTP error (%d); retrying in %g seconds.",
                     status, sleep_timer)
                 time.sleep(sleep_timer)
                 continue
             elif status == 429:
-                sleep_timer *= self.sleep_timer_base
+                sleep_timer *= self.cooldown_factor()
                 self.log.debug("Hit API rate limit (response status 429); "
                     "retrying in %g seconds", sleep_timer)
                 time.sleep(sleep_timer)
                 continue
             elif status == 401:
                 # Stop. Authentication failed. We shouldn't try doing any more,
-                # because we'll run into problems later anyway.
-                raise PDClientError(
-                    "Received 401 Unauthorized response from the API. The "
-                    "access key (%s) might not be valid."%self.trunc_key,
-                    response=response)
+                # because we'll run into the same problem later anyway.
+                raise PDHTTPError(
+                    "Received 401 Unauthorized response from the API. The key "
+                    "(...%s) may be invalid or deactivated."%self.trunc_key,
+                    response)
             else:
                 # All went according to plan.
                 return response
@@ -659,13 +670,18 @@ class EventsAPISession(PDSession):
         """
         return self.send_event('acknowledge', dedup_key=dedup_key)
 
-    def prepare_headers(self, method):
-        """Add user agent and content type headers for Events API requests."""
-        headers = deepcopy(self.headers)
+    def prepare_headers(self, method, user_headers={}):
+        """Add user agent and content type headers for Events API requests.
+
+        :param user_headers: User-supplied headers that will override defaults
+        """
+        headers = {}
+        headers.update(self.headers)
         headers.update({
             'Content-Type': 'application/json',
             'User-Agent': self.user_agent,
         })
+        headers.update(user_headers)
         return headers
 
     def resolve(self, dedup_key):
@@ -689,7 +705,7 @@ class EventsAPISession(PDSession):
         :param dedup_key:
             The deduplication key; used for determining event uniqueness and
             associating actions with existing incidents.
-        :param \*\*properties:
+        :param \\*\\*properties:
             Additional properties to set, i.e. if ``action`` is ``trigger``
             this would include ``payload``
         :type action: str
@@ -705,7 +721,7 @@ class EventsAPISession(PDSession):
         event = {'event_action':action}
 
         event.update(properties)
-        if isinstance(dedup_key, string_types):
+        if isinstance(dedup_key, str):
             event['dedup_key'] = dedup_key
         elif not action == 'trigger':
             raise ValueError("The dedup_key property is required for"
@@ -817,13 +833,15 @@ class ChangeEventsAPISession(PDSession):
     def event_timestamp(self):
         return datetime.utcnow().isoformat()+'Z'
 
-    def prepare_headers(self, method):
+    def prepare_headers(self, method, user_headers={}):
         """Add user agent and content type headers for Change Events API requests."""
         headers = deepcopy(self.headers)
         headers.update({
             'Content-Type': 'application/json',
             'User-Agent': self.user_agent,
         })
+        if user_headers:
+            headers.update(user_headers)
         return headers
 
     def send_change_event(self, **properties):
@@ -832,7 +850,7 @@ class ChangeEventsAPISession(PDSession):
 
         See: https://developer.pagerduty.com/docs/events-api-v2/send-change-events/
 
-        :param \*\*properties:
+        :param \\*\\*properties:
             Properties to set, i.e. ``payload`` and ``links``
         :returns:
             The response ID
@@ -942,6 +960,33 @@ class APISession(PDSession):
     def after_set_api_key(self):
         self._subdomain = None
 
+    @property
+    def api_key_access(self):
+        """
+        Memoized API key access type getter.
+
+        Will be "user" if the API key is a user-level token (all users should
+        have permission to create an API key with the same permissions as they
+        have in the PagerDuty web UI).
+
+        If the API key in use is an account-level API token (as only a global
+        administrator user can create), this property will be "account".
+        """
+        if not hasattr(self, '_api_key_access') or self._api_key_access is None:
+            response = self.get('/users/me')
+            if response.status_code == 400:
+                message = try_decoding(response).get('error', '')
+                if 'account-level access token' in message:
+                    self._api_key_access = 'account'
+                else:
+                    self._api_key_access = None
+                    self.log.error("Failed to obtain API key access level; "
+                        "the API did not respond as expected.")
+                    self.log.debug("Body = %s", response.text[:99])
+            else:
+                self._api_key_access = 'user'
+        return self._api_key_access
+
     @property
     def auth_type(self):
         """
@@ -1050,8 +1095,7 @@ class APISession(PDSession):
         Each yielded value is a dict object representing a result returned from
         the index. For example, if requesting the ``/users`` endpoint, each
         yielded value will be an entry of the ``users`` array property in the
-        response; see: `List Users
-        <https://v2.developer.pagerduty.com/v2/page/api-reference#!/Users/get_users>`_
+        response.
 
         :param path:
             The index endpoint URL to use.
@@ -1129,13 +1173,13 @@ class APISession(PDSession):
                     raise PDClientError("Encountered HTTP error status (%d) "
                         "response while iterating through index endpoint %s."%(
                             r.status_code, path), response=r)
-                self.log.debug("Stopping iteration on endpoint \"%s\"; API "
+                self.log.warn("Stopping iteration on endpoint \"%s\"; API "
                     "responded with non-success status %d", path, r.status_code)
                 break
             try:
                 response = r.json()
             except ValueError:
-                self.log.debug("Stopping iteration on endpoint \"%s\"; API "
+                self.log.warn("Stopping iteration on endpoint \"%s\"; API "
                     "responded with invalid JSON.", path)
                 break
             #if 'limit' in response:
@@ -1147,14 +1191,14 @@ class APISession(PDSession):
                 if 'more' in response:
                     more = response['more']
                 else:
-                    self.log.debug("Pagination is enabled in iteration, but the"
+                    self.log.warn("Pagination is enabled in iteration, but the"
                         " index endpoint %s responded with no \"more\" property"
                         " in the response. Only the first page of results, "
                         "however many can be gotten, will be included.", path)
                 if 'total' in response:
                     total_count = response['total']
                 else:
-                    self.log.debug("Pagination and the \"total\" parameter "
+                    self.log.warn("Pagination and the \"total\" parameter "
                         "are enabled in iteration, but the index endpoint %s "
                         "responded with no \"total\" property in the response. "
                         "Cannot display a total count of this resource without "
@@ -1167,6 +1211,50 @@ class APISession(PDSession):
                     item_hook(result, n, total_count)
                 yield result
 
+    def iter_cursor(self, path, attribute=None, params=None):
+        """
+        Iterator for results from an endpoint supporting cursor-based pagination
+
+        :param path:
+            The index endpoint URL to use.
+        :param attribute:
+            The property in the JSON body, i.e. ``records`` for
+            ``/audit/records``, in which to find results. If None is specified,
+            a guess will be made as to the envelope name (that it is the same as
+            the last node in the URI path).
+        :param params:
+            Query parameters to include in the request.
+        """
+
+        if attribute:
+            assumed_attribute = attribute
+        else:
+            assumed_attribute = path.split('/')[-1]
+
+        user_params = {}
+        if params:
+            user_params.update(params)
+
+        more = True
+        next_cursor = None
+        while more:
+            if next_cursor:
+                user_params.update({'cursor': next_cursor})
+            page = self.jget(path, params=user_params)
+            if assumed_attribute not in page:
+                raise PDClientError("No attribute \"{attribute}\" at the root "
+                    "level of the response body from {path}. Attributes "
+                    "include: {examples}".format(
+                        attribute = assumed_attribute,
+                        path = path,
+                        examples = (', '.join(page.keys()))[:99]
+                    )
+                )
+            for result in page[assumed_attribute]:
+                yield result
+            next_cursor = page.get('next_cursor', None)
+            more = bool(next_cursor)
+
     @auto_json
     def jget(self, path, **kw):
         """
@@ -1206,9 +1294,9 @@ class APISession(PDSession):
         """
         return list(self.iter_all(path, **kw))
 
-    def persist(self, resource, attr, values):
+    def persist(self, resource, attr, values, update=False):
         """
-        Finds or creates and returns a resource matching an idempotency key.
+        Finds or creates and returns a resource with a matching attribute
 
         Given a resource name, an attribute to use as an idempotency key and a
         set of attribute:value pairs as a dict, create a resource with the
@@ -1226,9 +1314,13 @@ class APISession(PDSession):
             The content of the resource to be created, if it does not already
             exist. This must contain an item with a key that is the same as the
             ``attr`` argument.
+        :param update:
+            (New in 4.4.0) If set to True, any existing resource will be updated
+            with the values supplied.
         :type resource: str
         :type attr: str
         :type values: dict
+        :type update: bool
         :rtype: dict
         """
         if attr not in values:
@@ -1236,8 +1328,15 @@ class APISession(PDSession):
                 "to the `attr` argument (expected idempotency key: '%s')."%attr)
         existing = self.find(resource, values[attr], attribute=attr)
         if existing:
+            if update:
+                original = {}
+                original.update(existing)
+                existing.update(values)
+                if original != existing:
+                    existing = self.rput(existing, json=existing)
             return existing
-        return self.rpost(resource, json=values)
+        else:
+            return self.rpost(resource, json=values)
 
     def postprocess(self, response, suffix=None):
         """
@@ -1246,10 +1345,9 @@ class APISession(PDSession):
         This method is called automatically by :func:`request` for all requests,
         and can be extended in child classes.
 
-        :param method:
-            Method of the request
         :param response:
-            Response object
+            The `requests.Response`_ object returned by
+            `requests.Session.request`_
         :param suffix:
             Optional suffix to append to the key
         :type method: str
@@ -1279,13 +1377,15 @@ class APISession(PDSession):
                 "and reference x_request_id=%s / date=%s",
                 status, request_id, request_date)
 
-    def prepare_headers(self, method):
+    def prepare_headers(self, method, user_headers={}):
         headers = deepcopy(self.headers)
         headers['User-Agent'] = self.user_agent
         if self.default_from is not None:
             headers['From'] = self.default_from
         if method in ('POST', 'PUT'):
             headers['Content-Type'] = 'application/json'
+        if user_headers:
+            headers.update(user_headers)
         return headers
 
     def profiler_key(self, method, path, suffix=None):
@@ -1320,7 +1420,7 @@ class APISession(PDSession):
             The path/URL to which to send the request, or a dict object
             representing an API resource that contains an item with key ``self``
             whose value is the URL of the resource.
-        :param \*\*kw:
+        :param \\*\\*kw:
             Keyword arguments to pass to ``requests.Session.delete``
         :type path: str or dict
         """
@@ -1330,13 +1430,13 @@ class APISession(PDSession):
     @resource_envelope
     def rget(self, resource, **kw):
         """
-        Retrieve a resource and return the encapsulated object in the response
+        Retrieve a resource and return the wrapped entity in the response
 
         :param resource:
             The path/URL to which to send the request, or a dict object
             representing an API resource that contains an item with key ``self``
             whose value is the URL of the resource.
-        :param \*\*kw:
+        :param \\*\\*kw:
             Keyword arguments to pass to ``requests.Session.rget``
         :returns:
             Dictionary representation of the object.
@@ -1356,7 +1456,7 @@ class APISession(PDSession):
         :param path:
             The path/URL to which to send the POST request, which should be an
             index endpoint.
-        :param \*\*kw:
+        :param \\*\\*kw:
             Keyword arguments to pass to ``requests.Session.post``
         :returns:
             Dictionary representation of the created object
@@ -1369,13 +1469,13 @@ class APISession(PDSession):
     @resource_envelope
     def rput(self, resource, **kw):
         """
-        Update an individual resource, returning the encapsulated object.
+        Update an individual resource, returning the wrapped entity as a dict
 
         :param resource:
             The path/URL to which to send the request, or a dict object
             representing an API resource that contains an item with key ``self``
             whose value is the URL of the resource.
-        :param \*\*kw:
+        :param \\*\\*kw:
             Keyword arguments to pass to ``requests.Session.put``
         :returns:
             Dictionary representation of the updated object
@@ -1430,3 +1530,42 @@ class PDClientError(Exception):
         self.msg = message
         self.response = response
         super(PDClientError, self).__init__(message)
+
+class PDHTTPError(PDClientError):
+    """
+    Error class representing errors strictly associated with HTTP responses.
+
+    This class was created to make it easier to more cleanly handle errors by
+    way of a class that is guaranteed to have its ``response`` be a valid
+    `requests.Response`_ object.
+
+    Whereas, the more generic :class:`PDClientError` could also be used
+    to denote such things as non-transient network errors wherein no response
+    was recevied from the API.
+
+    For instance, instead of this:
+
+    ::
+
+        try:
+            user = session.rget('/users/PABC123')
+        except pdpyras.PDClientError as e:
+            if e.response is not None:
+                print("HTTP error: "+str(e))
+            else:
+                print(e)
+
+    one could write this:
+
+    ::
+
+        try:
+            user = session.rget('/users/PABC123')
+        except pdpyras.PDHTTPErrror as e:
+            print("HTTP error: "+str(e))
+        except pdpyras.PDClientError as e:
+            print(e)
+    """
+
+    def __init__(self, message, response: requests.Response):
+        super(PDHTTPError, self).__init__(message, response)
